#include <stdio.h>
#include "zen.h"
#include "zenfact.h"


/* \TEX{
   \title{Setting a polynomial to an irreducible random one}
   \input{A polynomial, a degree, a |ZENRing|, and a function pointer on the algorithm to use (|Berlekamp| and |DDF| are available: see |ZENPolyIsIrreducible|).}
   \output{|ZENERR| if an error occured, 0 otherwise}
   \side{The polynomial is set randomly to a monic polynomial of
   given degree that is irreducible over |R|}
   \note{This procedure may take a while}
   }
   */
int _ZENPolySetRandomIrreducible(P,deg,Algo,R)
     ZENPoly P;
     int deg;
     int (*Algo) ___((ZENPoly, ZENRing));
     ZENRing R;
{
  int i;
  ZENPoly PX;

  if(ZENPolyAlloc(PX,deg,R)<0)
    return(ZERR);
      
  for(i=100;;i--)
    {
      ZENPolySetRandom(PX,deg,R);
      
      switch(_ZENPolyIsNotIrreducible(PX,Algo,R))
	{
	case ZENPOLY_IS_IRREDUCIBLE:
	  if(ZENPolyMakeMonic(P,PX,R)<0)
	    {
	      ZENPolyFree(PX,R);
	      return(ZERR);
	    }
	  ZENPolyFree(PX,R);
	  return(0);
	case ZENERR:
	  ZENPolyFree(PX,R);
	  return(ZERR);	 
	}
#if 0
# if 0
      if(i<0)
#endif
	{
	  fprintf(stderr,"+");fflush(stderr);
	}
# if 0
      if(i<0) i=100;
# endif
#endif
     
    }
}


/*
  \TEX{
  \title{Finding the smallest polynomial $f(X)$ such that $X^d+f(X)$ is an irreducible polynomial.}
  \input{A degree |d|, an allocated polynomial |P| of size at
  least |d| of a |ZENRing| |R|, and a function pointer on the algorithm to use (|Berlekamp| and |DDF| are available: see |ZENPolyIsIrreducible|).}
  \output{-1 if an error occured, 0 otherwise}
  \side{The polynomial is set to a polynomial of given degree that is irreducible over |R|}
  \note{This procedure may take a while.}
  } 
*/
int _ZENPolySetSmallestIrreducible(P,deg,Algo,R)
     ZENPoly P;
     int deg;
     int (*Algo) ___((ZENPoly, ZENRing));
     ZENRing R;
{
  int i;
  ZENElt elt;

  ZENPolySetToXi(P, deg, R);
  for(;;) {
    for(i = 0; i < deg; i++) {
      elt = ZENPolyGetCoeffPtr(P, i, R);
      ZENEltSetNext(elt, R);
      ZENPolySetCoeff(P, i, elt, R);
      if (!ZENEltIsZero(elt, R)) {
	break;
      }
#if 0
      fprintf(stderr,"P:"); ZENPolyPrintToFile(stderr,P,10,R); fprintf(stderr,"\n");
#endif
    }
    if (i == deg) {
      break;
    }

    switch(_ZENPolyIsNotIrreducible(P,Algo,R)) {
    case ZENPOLY_IS_IRREDUCIBLE:
      return(0);
    case ZENERR:
      return(ZERR);
    }
  }
  return(ZERR);
}

/*
  \TEX{
  \title{Finding a small random polynomial $f(X)$ such that $X^d+f(X)$ is an irreducible polynomial.}
  \input{A degree |d|, an allocated polynomial |P| of size at
  least |d| of a |ZENRing| |R|, and a function pointer on the algorithm to use (|Berlekamp| and |DDF| are available: see |ZENPolyIsIrreducible|).}
  \output{-1 if an error occured, 0 otherwise}
  \side{The polynomial is set to a polynomial of given degree that is irreducible over |R|}
  \note{This procedure may take a while.}
  } 
*/
int _ZENPolySetGoodIrreducible(P,deg,Algo,R)
     ZENPoly P;
     int deg;
     int (*Algo) ___((ZENPoly, ZENRing));
     ZENRing R;
{
  int i,subdeg;
  ZENPoly f;

  for(subdeg=1;subdeg<deg;subdeg++)
    {
      if(ZENPolyAlloc(f,subdeg,R)) {
	return(ZERR);
      }
      
      for(i=100;i--;)
	{
	  ZENPolySetRandom(f,subdeg,R);
	  ZENPolySetToXi(P, deg, R);	  
	  ZENPolyAdd(P,f,R);

	  switch(_ZENPolyIsNotIrreducible(P,Algo,R))
	    {
	    case ZENPOLY_IS_IRREDUCIBLE:
	      ZENPolyFree(f,R);
	      return(0);
	    case ZENERR:
	      ZENPolyFree(f,R);
	      return(ZERR);	 
	    }
	}
#if 0
      fprintf(stderr,"+"); fflush(stderr);
#endif
    }  
  return(_ZENPolySetRandomIrreducible(P,deg,Algo,R));
}
