#include <stdio.h>
#include "zen.h"
#include "zenfact.h"

/* \TEX{
   These functions aim at finding the roots of any polynomial $P$. The main
   function is |ZENPolyRoots| that calls the function
   |ZENPolyRootsBerlekamp| if its degree is greater than 2 or
   |ZENPolyRootsDegree2| otherwise.
   } */

#if 0
/*
   \TEX{
   \title{Roots of any polynomial.}
   \input{Two |ZENPoly|, |P| and |roots|, over a |ZENRing| |Rg|.}
   \output{\begin{itemize}
   \item[|ZENERR|] if an error occured,
   \item[|ZEN_NO_INVERSE|] if an inverse was impossible to compute,
   \item[|ZEN_HAS_INVERSE|] if the polynomial has one root.
   \end{itemize}}
   \side{The polynomial |roots| is set to a polynomial of degree -1,
   0,\ldots or degree(|P|)-1, the coefficients of which are the roots
   of |P|.} 
   \note{|roots| must be allocated for at least degree of |P| minus
   one. We must have $|P|\neq |roots|$. Valid only in finite fields.}
   } 
   */
int ZENPolyRoots(roots, P, Rg)
     ZENPoly roots, P;
     ZENRing Rg;{}
#endif
int _ZENPolyRoots(roots, P, Rg)
     ZENPoly roots, P;
     ZENRing Rg;
{
  ZENElt e, Xpq;
  ZENPoly Xq, gcd;
  ZENRing KRg;
  BigNum n;
  BigNumLength nl;
  int ret;
  ZENPrc prc;
  
  switch(ZENPolyDeg(P, Rg)) {
  case 0:
    ZENPolyDeg(roots, Rg)= -1; 
    return(ZEN_HAS_INVERSE);
  case 1:
    return(ZENPolyRootsDegree1(roots, P, Rg));
  case 2:
    return(ZENPolyRootsDegree2(roots, P, Rg));
  }

  ZENPolyDeg(roots, Rg) = -1;
  if ((ZENRingQl(Rg) == 1) &&
      (ZBNDigit(ZENRingQ(Rg), 0) == 2)) {
    ZENElt elt;

    elt = ZENPolyGetCoeffPtr(roots, 0, Rg);
    ZENEltSetToZero(elt, Rg);
    ZENPolyEval(elt, P, elt, Rg);
    if (ZENEltIsZero(elt, Rg)) {
      ZENPolyDeg(roots, Rg)++;
    }

    elt = ZENPolyGetCoeffPtr(roots, 1, Rg);
    ZENEltSetToOne(elt, Rg);
    ZENPolyEval(elt, P, elt, Rg);
    if (ZENEltIsZero(elt, Rg)) {
      ZENEltSetToOne(elt, Rg);
      ZENPolyDeg(roots, Rg)++;
      ZENPolySetCoeff(roots, ZENPolyDeg(roots, Rg), elt, Rg);
    }
    return(ZEN_HAS_INVERSE);
  }
  

  KRg = ZENExtRingCreate(P, Rg);
  CheckReturn(!KRg, ;, return(ZERR));
#define mbf_KRg ZENRingClose(KRg)
  if (ZENPrcIsSet(ZENRingPrcp(Rg), ZENPRC_ELT_MULTIPLY)) {
    ZENPrcSetNone(prc); 
    ZENPrcSet(prc, ZENPRC_ELT_MULTIPLY);
    ZENRingAddPrc(KRg, prc);
  }

  ZENEltAT(e, KRg, mbf_KRg, return(ZERR));
#define mbf_e ZENEltFree(e, KRg); mbf_KRg

  ZENEltAT(Xpq, KRg, mbf_e, return(ZERR));
#define mbf_Xpq ZENEltFree(Xpq, KRg); mbf_e

  ZENZToElt(e, ZENRingQ(Rg), ZENRingQl(Rg), KRg);
  ZENEltExp(Xpq, ZENRingQ(Rg), ZENRingQl(Rg), e, KRg);
  ZENEltSubtract(e, Xpq, KRg);

  if (!ZENEltIsZero(e, KRg)) {

    ZBNAT(n, ZENRingQl(KRg), mbf_Xpq, return(ZERR));
    ZENEltToZ(n, &nl, e, KRg);
    Xq  = ZENZToPoly(n, nl, Rg);
    ZBNF(n); mbf_Xpq;
    CheckReturn(!Xq, mbf_Xpq, return(ZERR));

    ZENPolyAT(gcd, ZENPolyDeg(P, Rg), Rg, ZENPolyFree(Xq, Rg), return(ZERR));
    CheckReturn((ret = ZENPolyGcd(gcd, Xq, P, Rg)) != ZEN_HAS_INVERSE,
		ZENPolyFree(Xq, Rg); ZENPolyFree(gcd, Rg),
		return(ret));
    ZENPolyFree(Xq, Rg);
    ret = ZENPolyRootsEDF(roots, gcd, Rg);
    ZENPolyFree(gcd, Rg);
  }
  else {
    mbf_Xpq;
    ret = ZENPolyRootsEDF(roots, P, Rg);
  }  
  return(ret);
}

#if 0
/*
   \TEX{
   \title{Roots of a polynomial whose degree is equal to its number of roots.}
   \input{Two |ZENPoly|, |P| and |roots|, over a |ZENRing| |Rg|.}
   \output{\begin{itemize}
   \item[|ZENERR|] if an error occured,
   \item[|ZEN_NO_INVERSE|] if an inverse was impossible to compute,
   \item[|ZEN_HAS_INVERSE|] if the polynomial has one root.
   \end{itemize}}
   \side{The polynomial |roots| is set to a polynomial of degree -1,
   0,\ldots or degree(|P|)-1, the coefficients of which are the roots
   of |P|.} 
   \note{|roots| must be allocated for at least degree of |P| minus
   one. We must have $|P|\neq |roots|$. Valid only in finite fields.}
   } 
   */
int ZENPolyRootsEDF(roots, P, Rg)
     ZENPoly roots, P;
     ZENRing Rg;{}
#endif
int _ZENPolyRootsEDF(roots, P, Rg)
     ZENPoly roots, P;
     ZENRing Rg;
{
  ZENPolyDeg(roots, Rg) = -1;
  if ((ZENRingPl(Rg) == 1) && (ZBNDigit(ZENRingP(Rg), 0) == 2)) {

    /* Characteristic 2 */
    return(ZENPolyRootsBerlekampEvenChar(roots, P, Rg));
  }

  /* Odd Characteristic */
  return(ZENPolyRootsBerlekampOddChar(roots, P, Rg));
}

#if 0
/*
   \TEX{
   \title{Roots of any polynomial in finite fields of even characteristic.}
   \input{Two |ZENPoly|, |P| and |roots|, over a |ZENRing| |Rg|.}
   \output{\begin{itemize}
   \item[|ZENERR|] if an error occured,
   \item[|ZEN_NO_INVERSE|] if an inverse was impossible to compute,
   \item[|ZEN_HAS_INVERSE|] if the polynomial has one root.
   \end{itemize}}
   \side{The polynomial |roots| is set to a polynomial of degree -1,
   0,\ldots or degree(|P|)-1, the coefficients of which are the roots
   of |P|.} 
   \note{|roots| must be allocated for at least degree of |P| minus
   one. We must have $|P|\neq |roots|$. Valid only in finite fields.}
   } 
   */
int ZENPolyRootsBerlekampEvenChar(roots, P, Rg)
     ZENPoly roots, P;
     ZENRing Rg;{}
#endif
int _ZENPolyRootsBerlekampEvenChar(roots, P, Rg)
     ZENPoly roots, P;
     ZENRing Rg;
{
  int AbsDeg;

  int i, j, ret;
  ZENPoly *PowerOfXModP, S, R;
  ZENElt Ke, beta, elt;
  BigNum tmp;
  BigNumLength tmpl;
  ZENRing rg, KRg;
  ZENPrc prc;

  switch(ZENPolyDeg(P, Rg)) {
  case 0:
    ZENPolyDeg(roots, Rg)= -1; 
    return(ZEN_HAS_INVERSE);
  case 1:
    return(ZENPolyRootsDegree1(roots, P, Rg));
  case 2:
    return(ZENPolyRootsDegree2(roots, P, Rg));
  }

  /* The absolute degree of Rg */
  AbsDeg = 1;
  for(rg = Rg; ZENRingDef(rg); rg = ZENRingDef(rg)) {
    AbsDeg *= ZENRingDeg(rg);
  }

  /* The extension defined by P */
  KRg = ZENExtRingCreate(P, Rg);
  CheckReturn(!KRg, ;, return(ZERR));
#define mbf_EC1 ZENRingClose(KRg)
  if (ZENPrcIsSet(ZENRingPrcp(Rg), ZENPRC_ELT_MULTIPLY)) {
    ZENPrcSetNone(prc); 
    ZENPrcSet(prc, ZENPRC_ELT_MULTIPLY);
    ZENRingAddPrc(KRg, prc);
  }
  
  /* x, x^2, ..., x^(2^AbsDeg) mod P(x)*/
  ZBNAT(tmp, ZENRingQl(KRg), mbf_EC1, return(ZERR));
  ZENEltAT(Ke, KRg, ZBNF(tmp); mbf_EC1, return(ZERR));
#define mbf_EC2 ZENEltFree(Ke, KRg); ZBNF(tmp); mbf_EC1
  MallocAT(PowerOfXModP, AbsDeg, ZENPoly, mbf_EC2, return(ZERR));
  ZENZToElt(Ke, ZENRingQ(Rg), ZENRingQl(Rg), KRg);
  PowerOfXModP[0] = ZENZToPoly(ZENRingQ(Rg), ZENRingQl(Rg), Rg);
  CheckReturn(!PowerOfXModP[0], ZENFree(PowerOfXModP);  mbf_EC2, return(ZERR));
  for(i = 1; i < AbsDeg; i++) {
    ZENEltSquare(Ke, Ke, KRg);
    ZENEltToZ(tmp, &tmpl, Ke, KRg);
    PowerOfXModP[i] = ZENZToPoly(tmp, tmpl, Rg);
    CheckReturn(!PowerOfXModP[i],
		for(i--; i >= 0; i--) {ZENPolyFree(PowerOfXModP[i], Rg);} ZENFree(PowerOfXModP);  mbf_EC2,
		return(ZERR));
  }
#define mbf_EC3 for(i = AbsDeg-1; i >= 0; i--) {ZENPolyFree(PowerOfXModP[i], Rg);} ZENFree(PowerOfXModP);  mbf_EC2

  ZENPolyAT(S, ZENPolyDeg(P, Rg), Rg, mbf_EC3, return(ZERR));
#define mbf_EC4 ZENPolyFree(S, Rg); mbf_EC3
  ZENPolyAT(R, ZENPolyDeg(P, Rg), Rg, mbf_EC4, return(ZERR));
#define mbf_EC5 ZENPolyFree(R, Rg); mbf_EC4
  ZENEltAT(beta, Rg, mbf_EC5, return(ZERR));
#define mbf_EC7 ZENEltFree(beta, Rg); mbf_EC5

  for(;;) {
    ZENEltSetRandom(beta, Rg);
    ZENPolySetToZero(S, Rg);
    for(j = 0; j < AbsDeg; j++) {
      if (j) {ZENEltSquare(beta, beta, Rg);}
      ZENPolyAssign(R, PowerOfXModP[j], Rg);
      ZENPolyMultiplyScalar(R, R, beta, Rg);
      ZENPolyAdd(S, R, Rg);
    }
    if (!ZENPolyDeg(S, Rg)) {continue;}
    ZENPolyMakeMonic(S, S, Rg);
    CheckReturn((ret = ZENPolyGcd(R, P, S, Rg)) != ZEN_HAS_INVERSE, mbf_EC7, return(ret));
    if (ZENPolyDeg(R, Rg)) {break;}

    elt = ZENPolyGetCoeffPtr(S, 0, Rg);
    ZENEltSetToOne(beta, Rg);
    ZENEltAdd(elt, beta, Rg);
    ZENPolySetCoeff(S, 0, elt, Rg);
    CheckReturn((ret = ZENPolyGcd(R, P, S, Rg)) != ZEN_HAS_INVERSE, mbf_EC7, return(ret));
    if (ZENPolyDeg(R, Rg)) {break;}
  }

  ZENPolyDeg(S, Rg) = -1;
  CheckReturn((ret = ZENPolyRootsBerlekampEvenChar(S, R, Rg)) != ZEN_HAS_INVERSE, mbf_EC7, return(ret));
  for(i = 0; i <= ZENPolyDeg(S, Rg); i++) {
    ZENPolySetCoeff(roots, ++ZENPolyDeg(roots, Rg), ZENPolyGetCoeffPtr(S, i, Rg), Rg);
  }

  CheckReturn((ret = ZENPolyDivide(S, R, P, R, Rg)) != ZEN_HAS_INVERSE, mbf_EC7, return(ret));

  ZENPolyDeg(R, Rg) = -1;
  CheckReturn((ret = ZENPolyRootsBerlekampEvenChar(R, S, Rg)) != ZEN_HAS_INVERSE, mbf_EC7, return(ret));
  for(i = 0; i <= ZENPolyDeg(R, Rg); i++) {
    ZENPolySetCoeff(roots, ++ZENPolyDeg(roots, Rg), ZENPolyGetCoeffPtr(R, i, Rg), Rg);
  }
  mbf_EC7;
  return(ZEN_HAS_INVERSE);
}

#if 0
/*
   \TEX{
   \title{Roots of any polynomial in finite fields of odd characteristic.}
   \input{Two |ZENPoly|, |P| and |roots|, over a |ZENRing| |Rg|.}
   \output{\begin{itemize}
   \item[|ZENERR|] if an error occured,
   \item[|ZEN_NO_INVERSE|] if an inverse was impossible to compute,
   \item[|ZEN_HAS_INVERSE|] if the polynomial has one root.
   \end{itemize}}
   \side{The polynomial |roots| is set to a polynomial of degree -1,
   0,\ldots or degree(|P|)-1, the coefficients of which are the roots
   of |P|.} 
   \note{|roots| must be allocated for at least degree of |P| minus
   one. We must have $|P|\neq |roots|$. Valid only in finite fields.}
   } 
   */
int ZENPolyRootsBerlekampOddChar(roots, P, Rg)
     ZENPoly roots, P;
     ZENRing Rg;{}
#endif
int _ZENPolyRootsBerlekampOddChar(roots, P, Rg)
     ZENPoly roots, P;
     ZENRing Rg;
{
  int i, ret;
  ZENPoly  S, R;
  ZENElt Ku, Kv, u, elt;
  BigNum tmp, e;
  BigNumLength tmpl, el;
  ZENRing KRg;
  ZENPrc prc;

  switch(ZENPolyDeg(P, Rg)) {
  case 0:
    ZENPolyDeg(roots, Rg)= -1; 
    return(ZEN_HAS_INVERSE);
  case 1:
    return(ZENPolyRootsDegree1(roots, P, Rg));
  case 2:
    return(ZENPolyRootsDegree2(roots, P, Rg));
  }

  /* The extension defined by P */
  KRg = ZENExtRingCreate(P, Rg);
  CheckReturn(!KRg, ;, return(ZERR));
#define mbf_OC1 ZENRingClose(KRg)
  if (ZENPrcIsSet(ZENRingPrcp(Rg), ZENPRC_ELT_MULTIPLY)) {
    ZENPrcSetNone(prc); 
    ZENPrcSet(prc, ZENPRC_ELT_MULTIPLY);
    ZENRingAddPrc(KRg, prc);
  }
  
  ZBNAT(tmp, ZENRingQl(KRg), mbf_OC1, return(ZERR));
  ZENEltAT(Kv, KRg, ZBNF(tmp); mbf_OC1, return(ZERR));
#define mbf_OC2 ZENEltFree(Kv, KRg); ZBNF(tmp); mbf_OC1

  ZENEltAT(Ku, KRg, mbf_OC2, return(ZERR));
#define mbf_OC3 ZENEltFree(Ku, KRg); mbf_OC2

  ZENEltAT(u, Rg, mbf_OC3, return(ZERR));
#define mbf_OC4 ZENEltFree(u, Rg); mbf_OC3

  ZENPolyAT(R, ZENPolyDeg(P, Rg), Rg, mbf_OC4, return(ZERR));
#define mbf_OC5 ZENPolyFree(R, Rg); mbf_OC4
 
  ZBNAT(e, ZENRingQl(Rg), mbf_OC5, return(ZERR));
#define mbf_OC6 ZENFree(e); mbf_OC5

  ZBNAssign(e, ZENRingQ(Rg), ZENRingQl(Rg));
  ZBNShiftRight(e, ZENRingQl(Rg), 1);
  el = ZBNNumDigits(e, ZENRingQl(Rg));

  ZENZToElt(Kv, ZENRingQ(Rg), ZENRingQl(Rg), KRg);
  for(;;) {
    ZENEltSetRandom(u, Rg);
    ZENEltToZ(tmp, &tmpl, u, Rg);
    ZENZToElt(Ku, tmp, tmpl, KRg);
    ZENEltAdd(Ku, Kv, KRg);

    ZENEltExp(Ku, e, el, Ku, KRg);

    ZENEltToZ(tmp, &tmpl, Ku, KRg);
    S = ZENZToPoly(tmp, tmpl, Rg);
    CheckReturn(!S, mbf_OC6, return(ZERR));

    ZENEltSetToOne(u, Rg);
    elt = ZENPolyGetCoeffPtr(S, 0, Rg);
    
    ZENEltAdd(elt, u, Rg);
    ZENPolySetCoeff(S, 0, elt, Rg);
    if (!ZENPolyIsZero(S, Rg)) {
      CheckReturn((ret = ZENPolyGcd(R, P, S, Rg)) != ZEN_HAS_INVERSE,
		  ZENPolyFree(S, Rg); mbf_OC6,
		  return(ret));
      if (ZENPolyDeg(R, Rg)) {break;}
    }

    ZENEltSubtract(elt, u, Rg);
    ZENEltSubtract(elt, u, Rg);
    ZENPolySetCoeff(S, 0, elt, Rg);
    if (!ZENPolyIsZero(S, Rg)) {
      CheckReturn((ret = ZENPolyGcd(R, P, S, Rg)) != ZEN_HAS_INVERSE,
		  ZENPolyFree(S, Rg); mbf_OC6,
		  return(ret));
      if (ZENPolyDeg(R, Rg)) {break;}
    }
    ZENPolyFree(S, Rg);
  }

  ZENPolyDeg(S, Rg) = -1;
  CheckReturn((ret = ZENPolyRootsBerlekampOddChar(S, R, Rg)) != ZEN_HAS_INVERSE, mbf_OC6, return(ret));
  for(i = 0; i <= ZENPolyDeg(S, Rg); i++) {
    ZENPolySetCoeff(roots, ++ZENPolyDeg(roots, Rg), ZENPolyGetCoeffPtr(S, i, Rg), Rg);
  }

  CheckReturn((ret = ZENPolyDivide(S, R, P, R, Rg)) != ZEN_HAS_INVERSE, mbf_OC6, return(ret));

  ZENPolyDeg(R, Rg) = -1;
  CheckReturn((ret = ZENPolyRootsBerlekampOddChar(R, S, Rg)) != ZEN_HAS_INVERSE, mbf_OC6, return(ret));
  for(i = 0; i <= ZENPolyDeg(R, Rg); i++) {
    ZENPolySetCoeff(roots, ++ZENPolyDeg(roots, Rg), ZENPolyGetCoeffPtr(R, i, Rg), Rg);
  }
  
  ZENPolyFree(S, Rg); mbf_OC6;
  return(ZEN_HAS_INVERSE);
}

#if 0
/*
   \TEX{
   \title{A root of a polynomial whose degree is equal to its number of roots.}
   \input{A |ZENPoly| |P| and an element |root| over a |ZENRing| |Rg|.}
   \output{\begin{itemize}
   \item[|ZENERR|] if an error occured,
   \item[|ZEN_NO_INVERSE|] if an inverse was impossible to compute,
   \item[|ZEN_HAS_INVERSE|] if the polynomial has one root.
   \end{itemize}}
   \side{|root| contains a root of |P|}
   \note{Valid only in finite fields.}
   } 
   */
int ZENPolyOneRootEDF(root, P, Rg)
     ZENElt root;
     ZENPoly P;
     ZENRing Rg;{}
#endif
int _ZENPolyOneRootEDF(p_root, P, Rg)
     ZENElt *p_root;
     ZENPoly P;
     ZENRing Rg;
{
  if ((ZENRingPl(Rg) == 1) && (ZBNDigit(ZENRingP(Rg), 0) == 2)) {

    /* Characteristic 2 */
    return(ZENPolyOneRootBerlekampEvenChar(*p_root, P, Rg));
  }

  /* Odd Characteristic */
  return(ZENPolyOneRootBerlekampOddChar(*p_root, P, Rg));
}

#if 0
/*
   \TEX{
   \title{A root of a polynomial whose degree is equal to its number
   of roots in finite fields of even characteristic.}
   \input{A |ZENPoly| |P| and an element |root| over a |ZENRing| |Rg|.}
   \output{\begin{itemize}
   \item[|ZENERR|] if an error occured,
   \item[|ZEN_NO_INVERSE|] if an inverse was impossible to compute,
   \item[|ZEN_HAS_INVERSE|] if the polynomial has one root.
   \end{itemize}}
   \side{|root| contains a root of |P|}
   \note{Valid only in finite fields.}
   } 
   */
int ZENPolyOneRootBerlekampEvenChar(root, P, Rg)
     ZENElt root;
     ZENPoly P;
     ZENRing Rg;{}
#endif
int _ZENPolyOneRootBerlekampEvenChar(p_root, P, Rg)
     ZENElt *p_root;
     ZENPoly P;
     ZENRing Rg;
{
  int AbsDeg;

  int i, j, ret;
  ZENPoly *PowerOfXModP, S, R;
  ZENElt Ke, beta, elt;
  BigNum tmp;
  BigNumLength tmpl;
  ZENRing rg, KRg;
  ZENPrc prc;

  if (ZENPolyDeg(P, Rg) == 1) {
    if ((ret = ZENEltInverse(*p_root, ZENPolyGetCoeffPtr(P, 1, Rg), Rg))
	!= ZEN_HAS_INVERSE) {
      return(ret);
    }
    ZENEltMultiply(*p_root, ZENPolyGetCoeffPtr(P, 0, Rg), *p_root, Rg);
    ZENEltNegate(*p_root, *p_root, Rg);
    return(ZEN_HAS_INVERSE);
  }
  if (ZENPolyDeg(P, Rg) == 2) {
    ZENPolyAT(S, ZENPolyDeg(P, Rg), Rg, ;, return(ZERR));
    if ((ret = ZENPolyRootsDegree2(S, P, Rg)) != ZEN_HAS_INVERSE) {
      ZENPolyFree(S, Rg);
      return(ret);
    }
    ZENEltAssign(*p_root, ZENPolyGetCoeffPtr(S, 0, Rg), Rg);
    ZENPolyFree(S, Rg);
    return(ZEN_HAS_INVERSE);
  }

  /* The absolute degree of Rg */
  AbsDeg = 1;
  for(rg = Rg; ZENRingDef(rg); rg = ZENRingDef(rg)) {
    AbsDeg *= ZENRingDeg(rg);
  }

  /* The extension defined by P */
  KRg = ZENExtRingCreate(P, Rg);
  CheckReturn(!KRg, ;, return(ZERR));
#define mbf_ECO1 ZENRingClose(KRg)
  if (ZENPrcIsSet(ZENRingPrcp(Rg), ZENPRC_ELT_MULTIPLY)) {
    ZENPrcSetNone(prc); 
    ZENPrcSet(prc, ZENPRC_ELT_MULTIPLY);
    ZENRingAddPrc(KRg, prc);
  }
  
  /* x, x^2, ..., x^(2^AbsDeg) mod P(x)*/
  ZBNAT(tmp, ZENRingQl(KRg), mbf_ECO1, return(ZERR));
  ZENEltAT(Ke, KRg, ZBNF(tmp); mbf_ECO1, return(ZERR));
#define mbf_ECO2 ZENEltFree(Ke, KRg); ZBNF(tmp); mbf_ECO1
  MallocAT(PowerOfXModP, AbsDeg, ZENPoly, mbf_ECO2, return(ZERR));
  ZENZToElt(Ke, ZENRingQ(Rg), ZENRingQl(Rg), KRg);
  PowerOfXModP[0] = ZENZToPoly(ZENRingQ(Rg), ZENRingQl(Rg), Rg);
  CheckReturn(!PowerOfXModP[0], ZENFree(PowerOfXModP);  mbf_ECO2, return(ZERR));
  for(i = 1; i < AbsDeg; i++) {
    ZENEltSquare(Ke, Ke, KRg);
    ZENEltToZ(tmp, &tmpl, Ke, KRg);
    PowerOfXModP[i] = ZENZToPoly(tmp, tmpl, Rg);
    CheckReturn(!PowerOfXModP[i],
		for(i--; i >= 0; i--) {ZENPolyFree(PowerOfXModP[i], Rg);} ZENFree(PowerOfXModP);  mbf_ECO2,
		return(ZERR));
  }
#define mbf_ECO3 for(i = AbsDeg-1; i >= 0; i--) {ZENPolyFree(PowerOfXModP[i], Rg);} ZENFree(PowerOfXModP);  mbf_ECO2

  ZENPolyAT(S, ZENPolyDeg(P, Rg), Rg, mbf_ECO3, return(ZERR));
#define mbf_ECO4 ZENPolyFree(S, Rg); mbf_ECO3
  ZENPolyAT(R, ZENPolyDeg(P, Rg), Rg, mbf_ECO4, return(ZERR));
#define mbf_ECO5 ZENPolyFree(R, Rg); mbf_ECO4
  ZENEltAT(beta, Rg, mbf_ECO5, return(ZERR));
#define mbf_ECO7 ZENEltFree(beta, Rg); mbf_ECO5

  for(i = 0; i < ZENRingDeg(KRg); i++) {
    ZENEltSetRandom(beta, Rg);
    ZENPolySetToZero(S, Rg);
    for(j = 0; j < AbsDeg; j++) {
      if (j) {ZENEltSquare(beta, beta, Rg);}
      ZENPolyAssign(R, PowerOfXModP[j], Rg);
      ZENPolyMultiplyScalar(R, R, beta, Rg);
      ZENPolyAdd(S, R, Rg);
    }
    if (!ZENPolyDeg(S, Rg)) {continue;}
    ZENPolyMakeMonic(S, S, Rg);
    CheckReturn((ret = ZENPolyGcd(R, P, S, Rg)) != ZEN_HAS_INVERSE, mbf_ECO7, return(ret));
    if (ZENPolyDeg(R, Rg)) {break;}

    elt = ZENPolyGetCoeffPtr(S, 0, Rg);
    ZENEltSetToOne(beta, Rg);
    ZENEltAdd(elt, beta, Rg);
    ZENPolySetCoeff(S, 0, elt, Rg);
    CheckReturn((ret = ZENPolyGcd(R, P, S, Rg)) != ZEN_HAS_INVERSE, mbf_ECO7, return(ret));
    if (ZENPolyDeg(R, Rg)) {break;}
  }

  if (ZENPolyDeg(R, Rg) > ZENPolyDeg(P, Rg)-ZENPolyDeg(R, Rg)) {    
    CheckReturn((ret = ZENPolyDivide(S, R, P, R, Rg)) != ZEN_HAS_INVERSE, mbf_ECO7, return(ret));
    ZENPolyAssign(R, S, Rg);
  }
  ret = ZENPolyOneRootBerlekampEvenChar(*p_root, R, Rg);
  mbf_ECO7;
  return(ret);
}

#if 0
/*
   \TEX{
   \title{A root of a polynomial whose degree is equal to its number
   of roots in finite fields of odd characteristic.}
   \input{A |ZENPoly| |P| and an element |root| over a |ZENRing| |Rg|.}
   \output{\begin{itemize}
   \item[|ZENERR|] if an error occured,
   \item[|ZEN_NO_INVERSE|] if an inverse was impossible to compute,
   \item[|ZEN_HAS_INVERSE|] if the polynomial has one root.
   \end{itemize}}
   \side{|root| contains a root of |P|}
   \note{Valid only in finite fields.}
   } 
   */
int ZENPolyOneRootBerlekampOddChar(root, P, Rg)
     ZENElt root;
     ZENPoly P;
     ZENRing Rg;{}
#endif
int _ZENPolyOneRootBerlekampOddChar(p_root, P, Rg)
     ZENElt *p_root;
     ZENPoly P;
     ZENRing Rg;
{
  int ret;
  ZENPoly  S, R;
  ZENElt Ku, Kv, u, elt;
  BigNum tmp, e;
  BigNumLength tmpl, el;
  ZENRing KRg;
  ZENPrc prc;

  if (ZENPolyDeg(P, Rg) == 1) {

    if ((ret = ZENEltInverse(*p_root, ZENPolyGetCoeffPtr(P, 1, Rg), Rg))
	!= ZEN_HAS_INVERSE) {
      return(ret);
    }
    ZENEltMultiply(*p_root, ZENPolyGetCoeffPtr(P, 0, Rg), *p_root, Rg);
    ZENEltNegate(*p_root, *p_root, Rg);
    return(ZEN_HAS_INVERSE);
  }

  if (ZENPolyDeg(P, Rg) == 2) {
    ZENPolyAT(S, ZENPolyDeg(P, Rg), Rg, ;, return(ZERR));
    if ((ret = ZENPolyRootsDegree2(S, P, Rg)) != ZEN_HAS_INVERSE) {
      ZENPolyFree(S, Rg);
      return(ret);
    }
    ZENEltAssign(*p_root, ZENPolyGetCoeffPtr(S, 0, Rg), Rg);
    ZENPolyFree(S, Rg);
    return(ZEN_HAS_INVERSE);
  }

  /* The extension defined by P */
  KRg = ZENExtRingCreate(P, Rg);
  CheckReturn(!KRg, ;, return(ZERR));
#define mbf_OCO1 ZENRingClose(KRg)
  if (ZENPrcIsSet(ZENRingPrcp(Rg), ZENPRC_ELT_MULTIPLY)) {
    ZENPrcSetNone(prc); 
    ZENPrcSet(prc, ZENPRC_ELT_MULTIPLY);
    ZENRingAddPrc(KRg, prc);
  }
  
  ZBNAT(tmp, ZENRingQl(KRg), mbf_OCO1, return(ZERR));
  ZENEltAT(Kv, KRg, ZBNF(tmp); mbf_OCO1, return(ZERR));
#define mbf_OCO2 ZENEltFree(Kv, KRg); ZBNF(tmp); mbf_OCO1

  ZENEltAT(Ku, KRg, mbf_OCO2, return(ZERR));
#define mbf_OCO3 ZENEltFree(Ku, KRg); mbf_OCO2

  ZENEltAT(u, Rg, mbf_OCO3, return(ZERR));
#define mbf_OCO4 ZENEltFree(u, Rg); mbf_OCO3

  ZENPolyAT(R, ZENPolyDeg(P, Rg), Rg, mbf_OCO4, return(ZERR));
#define mbf_OCO5 ZENPolyFree(R, Rg); mbf_OCO4
 
  ZBNAT(e, ZENRingQl(Rg), mbf_OCO5, return(ZERR));
#define mbf_OCO6 ZENFree(e); mbf_OCO5

  ZBNAssign(e, ZENRingQ(Rg), ZENRingQl(Rg));
  ZBNShiftRight(e, ZENRingQl(Rg), 1);
  el = ZBNNumDigits(e, ZENRingQl(Rg));

  ZENZToElt(Kv, ZENRingQ(Rg), ZENRingQl(Rg), KRg);
  for(;;) {
    ZENEltSetRandom(u, Rg);
    ZENEltToZ(tmp, &tmpl, u, Rg);
    ZENZToElt(Ku, tmp, tmpl, KRg);
    ZENEltAdd(Ku, Kv, KRg);

    ZENEltExp(Ku, e, el, Ku, KRg);

    ZENEltToZ(tmp, &tmpl, Ku, KRg);
    S = ZENZToPoly(tmp, tmpl, Rg);
    CheckReturn(!S, mbf_OCO6, return(ZERR));

    ZENEltSetToOne(u, Rg);
    elt = ZENPolyGetCoeffPtr(S, 0, Rg);
    
    ZENEltAdd(elt, u, Rg);
    ZENPolySetCoeff(S, 0, elt, Rg);
    if (!ZENPolyIsZero(S, Rg)) {
      CheckReturn((ret = ZENPolyGcd(R, P, S, Rg)) != ZEN_HAS_INVERSE,
		  ZENPolyFree(S, Rg); mbf_OCO6,
		  return(ret));
      if (ZENPolyDeg(R, Rg)) {break;}
    }

    ZENEltSubtract(elt, u, Rg);
    ZENEltSubtract(elt, u, Rg);
    ZENPolySetCoeff(S, 0, elt, Rg);
    if (!ZENPolyIsZero(S, Rg)) {
      CheckReturn((ret = ZENPolyGcd(R, P, S, Rg)) != ZEN_HAS_INVERSE,
		  ZENPolyFree(S, Rg); mbf_OCO6,
		  return(ret));
      if (ZENPolyDeg(R, Rg)) {break;}
    }
    ZENPolyFree(S, Rg);
  }
  if (ZENPolyDeg(R, Rg) > ZENPolyDeg(P, Rg)-ZENPolyDeg(R, Rg)) {    
    CheckReturn((ret = ZENPolyDivide(S, R, P, R, Rg)) != ZEN_HAS_INVERSE, ZENPolyFree(S, Rg); mbf_OCO6, return(ret));
    ZENPolyAssign(R, S, Rg);
  }
  ret = ZENPolyOneRootBerlekampOddChar(*p_root, R, Rg);
  ZENPolyFree(S, Rg); mbf_OCO6;
  return(ret);
}


int _ZENEltPRoot(p_a, b, Rg)
     ZENElt *p_a, b;
     ZENRing Rg;
{
  BigNum tmp;
  int tmpl;

  ZBNAT(tmp, tmpl = ZENRingQl(Rg)+1, ;, return(-1));
  ZBNSetToZero(tmp, tmpl);
  ZBNAssign(tmp, ZENRingQ(Rg), ZENRingQl(Rg));
  ZBNDivide(tmp, tmpl, ZENRingP(Rg), ZENRingPl(Rg));
  tmpl = ZBNNumDigits(tmp+ZENRingPl(Rg), 1+ZENRingQl(Rg)-ZENRingPl(Rg));
  ZENEltExp(*p_a, tmp+1, tmpl, b, Rg);
  ZBNF(tmp);
  return(0);
}

#if 0
/* \TEX{
   \title{$d$-roots in finite fields}.
   \input{Two elements |R| and |A| of a finite field |Rg|.}
   \output{|ZENERR| if an error occurred, 0 if |A| is a $d$-root, 1 otherwise.}
   \side{|R| is filled with the $d$-root of |A| if 0 is returned.}
   \note{This procedure is valid only in finite fields.}
   }
   */
int ZENEltDRoot(R, d, A, Rg)
     BigNumDigit d;
     ZENElt R, A;
     ZENRing Rg;{}
#endif
int _ZENEltDRoot(R, d, A, Rg)
     BigNumDigit d;
     ZENElt *R, A;
     ZENRing Rg;
{
  BigNum m, e1, e2, t;
  BigNumLength ml, e1l, e2l, tl;
  BigNumDigit d1, r;
  ZENRing rg;
  ZENElt g;
  ZENElt E, B, C;
  int ret, i;

  if (!(m = ZBNC(ZENRingQl(Rg)+1))) {
    return(ZERR);
  }
  if (!(e1 = ZBNC(ZENRingQl(Rg)+1))) {
    ZBNF(m);
    return(ZERR);
  }
  if (!(e2 = ZBNC(ZENRingQl(Rg)+1))) {
    ZBNF(m); ZBNF(e1);
    return(ZERR);
  }
  if (!(t = ZBNC(ZENRingQl(Rg)+1))) {
    ZBNF(m); ZBNF(e1);
    return(ZERR);
  }

  if (ZENEltCreate(E, Rg)) {
    ZBNF(m); ZBNF(e1); ZBNF(e2); ZBNF(t);
    return(ZERR);
  }
  if (ZENEltCreate(B, Rg)) {
    ZENEltFree(E, Rg);
    ZBNF(m); ZBNF(e1); ZBNF(e2); ZBNF(t);
    return(ZERR);
  }
  if (ZENEltCreate(C, Rg)) {
    ZENEltFree(E, Rg); ZENEltFree(B, Rg);
    ZBNF(m); ZBNF(e1); ZBNF(e2); ZBNF(t);
    return(ZERR);
  }

  /* m := q-1 */
  ZBNAssign(m, ZENRingQ(Rg), ZENRingQl(Rg));
  ZBNSubtractBorrow(m, ZENRingQl(Rg), 0);
  ml = ZBNNumDigits(m, ZENRingQl(Rg));
  ZBNDigit(m, ml) = 0;

  /* e1 := m/d */
  r = ZBNDivideDigit(e1, m, ml+1, d);
  if (r) {
    
    /* d doesn't divide m, easy, there exists no d-th root ! */
    ret = 1;
    goto end;
  }
  e1l = ZBNNumDigits(e1, ml);

  /* A generator of the d-th root of unity */
  for(;;) {
    for(ZENEltSetRandom(B, Rg);
	ZENEltIsZero(B, Rg);
	ZENEltSetRandom(B, Rg)) {;}
    ZENEltExp(E, e1, e1l, B, Rg);
    ZENEltSetToOne(C, Rg);
    for(i = 1; i < d; i++) {
      ZENEltMultiply(C, E, C, Rg);
      if (ZENEltIsOne(C, Rg)) {break;}
    }
    if (i == d) { break;}
  }

  /* d1 := gcd(e1, d) */
  ZBNDigit(e1, e1l) = 0;
  d1 = ZBNDivideDigit(e2, e1, e1l+1, d);
  d1 = ZBNDigitGCD(d1, d);

  /* e2 := 0 */
  ZBNSetToZero(e2, ZENRingQl(Rg)+1);
  e2l = 0;

  while(d1 != 1) {
    /* e1 := e1/d1 */
    ZBNDigit(e1, e1l) = 0;
    ZBNDivideDigit(e1, e1, e1l+1, d1);
    e1l = ZBNNumDigits(e1, e1l);
    ZBNSetToZero(e1+e1l, ZENRingQl(Rg)+1-e1l);

    /* e2 := e2/d1 */
    ZBNDigit(e2, e2l) = 0;
    ZBNDivideDigit(e2, e2, e2l+1, d1);
    e2l = ZBNNumDigits(e2, e2l);
    ZBNSetToZero(e2+e2l, ZENRingQl(Rg)+1-e2l);

    /* t := m/d1 */
    ZBNDivideDigit(t, m, ml+1, d1);
    tl = ZBNNumDigits(t, ml);

    /* while (A^e1 B^e2 <> 1) */
    ZENEltExp(C, e1, e1l, A, Rg);
    ZENEltExp(E, e2, e2l, B, Rg);
    ZENEltMultiply(C, E, C, Rg);
    while(!ZENEltIsOne(C, Rg)) {

      /* e2 := e2+t */
      ZENEltExp(E, t, tl, B, Rg);
      ZENEltMultiply(C, E, C, Rg);
      ZBNAdd(e2, ZENRingQl(Rg)+1, t, tl, 0);
      e2l = ZBNNumDigits(e2, ZENRingQl(Rg)+1);
    }

    /* d1 := gcd(e1, d) */
    ZBNDigit(e1, e1l) = 0;
    d1 = ZBNDivideDigit(t, e1, e1l+1, d);
    d1 = ZBNDigitGCD(d1, d);

  }

  /* r := -1/e1 mod d */
  if (!(rg = ZENBaseRingCreate(&d, 1))) {
    ret = -1;
    goto end;
  }
  if (ZENEltCreate(g, rg)) {
    ZENRingClose(rg);
    ret = -1;
    goto end;
  }
  ZBNDigit(e1, e1l) = 0;
  r = ZBNDivideDigit(t, e1, e1l+1, d);
  ZENZToElt(g, &r, 1, rg);
  ZENEltNegate(g, g, rg);
  ZENEltInverse(g, g, rg);
  ZENEltToZ(t, &tl, g, rg);
  r = ZBNDigit(t, 0);
  ZENEltFree(g, rg);
  ZENRingClose(rg);

  /* A^((g e1+1)/d) */
  ZBNSetToZero(t, ZENRingQl(Rg)+1);
  ZBNMultiplyDigit(t, ZENRingQl(Rg)+1, e1, e1l, r);
  ZBNAddCarry(t, ZENRingQl(Rg)+1, 1);
  ZBNDivideDigit(t, t, ZENRingQl(Rg)+1, d);
  tl = ZBNNumDigits(t, ZENRingQl(Rg)+1);
  ZENEltExp(*R, t, tl, A, Rg);
  
  /* B^((g e2)/d) */
  ZBNSetToZero(t, ZENRingQl(Rg)+1);
  ZBNMultiplyDigit(t, ZENRingQl(Rg)+1, e2, e2l, r);
  ZBNDivideDigit(t, t, ZENRingQl(Rg)+1, d);
  tl = ZBNNumDigits(t, ZENRingQl(Rg)+1);
  ZENEltExp(C, t, tl, B, Rg);
  
  ZENEltMultiply(*R, C, *R, Rg);
  
  ret = 0;
end:
  ZENEltFree(E, Rg); ZENEltFree(B, Rg); ZENEltFree(C, Rg);
  ZBNF(m); ZBNF(e1); ZBNF(e2); ZBNF(t);
  return(ret);
}

#if 0
/* \TEX{
   \title{Is a $d$-root in finite fields.}
   \input{An element |A| of a finite ring |Rg|.}
   \output{|ZENERR| if an error occurred, 1 if |A| is a $d$-root, 0 otherwise.}
   \note{This procedure is valid only in finite fields. The algorithm
   uses an exponantiation.}
   }
   */
int ZENEltIsADRoot(d, A, Rg)
     BigNumDigit d;
     ZENElt A;
     ZENRing Rg;{}
#endif
int _ZENEltIsADRoot(d, A, Rg)
     BigNumDigit d;
     ZENElt A;
     ZENRing Rg;
{
  BigNum m;
  int ml, ret;
  ZENElt R;

  /* Allocation */
  if (!(m = ZBNC(ZENRingQl(Rg)+1))) {
    return(ZERR);
  }
  if (ZENEltCreate(R, Rg)) {
    ZBNF(m);
    return(ZERR);
  }

  /* (q-1)/d */
  ZBNAssign(m, ZENRingQ(Rg), ZENRingQl(Rg));
  ZBNSubtractBorrow(m, ZENRingQl(Rg), 0);
  ml = ZBNNumDigits(m, ZENRingQl(Rg));
  ZBNDigit(m, ml) = 0;
  if ((ZBNDigit(m, 0) = ZBNDivideDigit(m+1, m, ml+1, d))) {
    ret = ZEN_HAS_INVERSE;
  }
  else {
    
    /* R := A^((q-1)/d) */
    ml = ZBNNumDigits(m+1, ml);
    ZENEltExp(R, m+1, ml, A, Rg);
    ret = ZENEltIsOne(R, Rg);
  }
  ZBNF(m);
  ZENEltFree(R, Rg);
  return(ret);
}
